%This is a script used to segment ImageField8 in 3D with the active mask
%algorithm to reproduce Fig. 8.
%
%       This file is part of the Matlab code provided for the following
%       reproducible paper:
%
%       Gowri Srinivasa, Matthew C. Fickus, Yusong Guo, Adam D. Linstedt, Jelena Kovacevic,
%       "Active mask segmentation of fluorescence microscope cell images",
%       IEEE Transactions on Image Processing. 
%                                                                           
%       This program is free software; you can redistribute it and/or modify it  
%       under the terms of the GNU General Public License as published by the     
%       Free Software Foundation; either version 2 of the License, or (at your    
%       option) any later version. This software is distributed in the hope that  
%       it will be useful, but without any warranty; without even the implied     
%       warranty of merchantability or fitness for a particular purpose.          
%       See the GNU General Public License for more details                       
%       (enclosed in the file COPYING).   
% 
%       GNU General Public License,
%       Copyright (C) 2008,
%       bimagicLab, Center for Bioimage Informatics (CBI)
%       Carnegie Mellon University,
%       Pittsburgh, PA 15213, USA.
%                                                                           
% COMMENTS
%
%       Authors: Gowri Srinivasa and Matthew C. Fickus
%       Latest modifications: April 22, 2008.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

format compact;
clear all;
close all;
clc;
TotTime=0;%measures the total time taken
xPad=16;
yPad=16;
zPad=16;
epsilon=0.0001;
bfactor=1/12; %introduced to be able to weight the scale based on the size of the image (like 'shrink' is used with 'dil')
b=(1/bfactor);%64 for the original image
c=1/8; %this can be annealed along with b to figure out the 'best fit' for a scale value;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Initialization of the masks; more the randomness, greater the chance of
%achieving a 'good cut'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
L=2;
initP=2^(2*L); %Upto 1600 masks work in 2D without running into cache problems

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Initialization of the parameters used in the (force) computations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
InScale=0.5;  %to dilate the error function used to compute the region force
OutScale=0.9; %same as lambda (wt. of Rforce) used in MSAC transform
Kmax=0; %dilation factor, 0 for the corasest version
shrink=3-Kmax; %shrink=0 for native resolution

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Creating a folder to store results along with the code and add the time stamp
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
current_time = datestr(now,'yyyy_mm_dd_HH_MM_SS');
filenamebase=sprintf('MultiMask_%s',current_time);
dos(sprintf('md %s',filenamebase));
dos(sprintf('copy AM3Dcode.m %s\\AM3Dcode.m',filenamebase));

%%%%%%%%%%%%%%%%%%%%%What is img.mat?%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%To be done ONCE for a stack. COMMENT IF NOT REQD.%%%%%%%%%%%%
%  img = zeros(128,168,40);
%  I = zeros(1024,1344);
%  level = 3;
%  path='C:\Stack_8\YSG2-21_T0002_WavelengthA_Z00';
%  nImages = 24;
%  for imno = 1:nImages
%      I = imread(sprintf('%s%02d.tif',path,imno));
%      T = wpdec2(double(I),level,'haar'); %level=3 levels of decomposition; filters not rotation or shift invariant
%      % extract the coarse subband (course approx = nodes 1,5,21)
%      % from the third level (node 21) of decomposition
%      img(:,:,imno) = wpcoef(T,21);
%  end
% 
% save 'img.mat' img;
% clear img I T;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
load img.mat;
[M,N,Q]=size(img);
imgMask = padarray(ones(size(img)),[xPad,yPad,zPad],0,'both');
img=padarray(img,[xPad,yPad,zPad],0,'both');
[Mpad,Npad,Qpad]=size(img);

[X,Y,Z]=ndgrid([0:floor(Mpad/2),-floor(Mpad/2)+1:-1],[0:floor(Npad/2),-floor(Npad/2)+1:-1],[0:floor(Qpad/2),-floor(Qpad/2)+1:-1]);
Z=(6/8)*Z; 

j=1;
dil=12;
a=dil(j);
g1=exp(-(1/(a^2))*(X.^2+Y.^2+Z.^2)); % gaussian window
g1Scale=ifftn(fftn(imgMask).*fftn(g1));
f=(ifftn(fftn(img).*fftn(g1)))./(g1Scale+epsilon);
f=f(xPad+1:end-xPad,yPad+1:end-yPad,zPad+1:end-zPad);
low=quantile(f(:),0.8);
high=quantile(f(:),0.95);

rscalar=-OutScale.*erf((2*f-low-high)/(InScale*(high-low)));

P=initP;
colorP=P;
colorIndex=1:P;
phi=zeros(M,N,Q,P); %multivalued function initlized at all zeros
psi=ceil(P*rand(M,N,Q)); %we introduce P numbers into the MXNXQ mask Psi
%psi=kron(reshape(1:2^(2*L),2^L,2^L),ones(2^(10-L-Kmax-shrink),2^(10-L-Kmax-shrink)));

for p=1:P
    temp=phi(:,:,:,p);
    temp(psi==p)=1; %an indicator function to identify where psi has the value p
    phi(:,:,:,p)=temp; %the values in the pth dimension are set to the indicator function (above)
end
[maxval, psi] = max(phi,[],4); %the updated psi is computed as the maxarg of phi (in the pth dimension)

g2=0.5*(1-erf((X.^2+Y.^2+Z.^2-b^2)/(b^2+c))); %smaller the "support", slower the computation
g2hat = fftn(g2);
g2Scale=ifftn(fftn(imgMask).*g2hat);


change=1;
iter=0;
phiPad=zeros(Mpad,Npad);
while change~=0
    tic;
    iter=iter+1;
    for p = 1:P
        phiPad=padarray(phi(:,:,:,p),[xPad,yPad,zPad],0,'both');
        phiPad=ifftn(fftn(phiPad).*g2hat)./(g2Scale+epsilon);
        phi(:,:,:,p)=phiPad(xPad+1:end-xPad,yPad+1:end-yPad,zPad+1:end-zPad);
    end
    phi(:,:,:,1) = phi(:,:,:,1)+rscalar;
    [maxval, inwhat] = max(phi,[],4);
    newPsi=zeros(M,N,Q);
    phi = zeros(M,N,Q,P);
    newP = 0;
    for p=1:P %as phi and psi evolve, defragment phi to get rid of space allotted to nonexisting dimensions
        temp = zeros(M,N,Q);
        temp(inwhat==p)=1;
        if(sum(sum(sum(temp)))~=0) %check if any dimension of phi has been wiped out
            newP = newP+1;
            phi(:,:,:,newP)=temp; %defragmenting the embedding function phi
            colorIndex(newP)=colorIndex(p); %as we get rid of non existing dimension, we need to reassign color values to keep the hues from changing
            newPsi = newPsi+colorIndex(p)*temp;
        end
    end
    P = newP;
    change=sum(sum(sum(((newPsi-psi)~=0)))); %how many pixels in the mask have moved?
    disp(sprintf('P=%03d; change=%05d; image %03d: Approximate time taken: %2.2f seconds',P,change,iter,toc));
    TotTime = TotTime+toc; %this is a very approximate measure; comment the above three lines to get a better estimate
    psi = newPsi;
end
beep;
save(strcat(filenamebase,'\\endState.mat')); %this saves the workspace
