%This is a script used to segment ImageField9 with the active mask
%algorithm to reproduce results reported in Table 1.
%
%       This file is part of the Matlab code provided for the following
%       reproducible paper:
%
%       Gowri Srinivasa, Matthew C. Fickus, Yusong Guo, Adam D. Linstedt, Jelena Kovacevic,
%       "Active mask segmentation of fluorescence microscope cell images",
%       IEEE Transactions on Image Processing. 
%                                                                           
%       This program is free software; you can redistribute it and/or modify it  
%       under the terms of the GNU General Public License as published by the     
%       Free Software Foundation; either version 2 of the License, or (at your    
%       option) any later version. This software is distributed in the hope that  
%       it will be useful, but without any warranty; without even the implied     
%       warranty of merchantability or fitness for a particular purpose.          
%       See the GNU General Public License for more details                       
%       (enclosed in the file COPYING).   
% 
%       GNU General Public License,
%       Copyright (C) 2008,
%       bimagicLab, Center for Bioimage Informatics (CBI)
%       Carnegie Mellon University,
%       Pittsburgh, PA 15213, USA.
%                                                                           
% COMMENTS
%
%       Authors: Gowri Srinivasa and Matthew C. Fickus
%       Latest modifications: April 22, 2008.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%function [psi-1, TotTime] = AMS(Kmax,low,high)
%function [psi-1, TotTime] = AMS(path,imgN, Kmax,low,high)
%Multiresolution Multiscale Active Mask Segmentation

%%%%%%%%%%%%%%%%%%Input%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Kmax - degree of refinement or number of MR levels of refining the contour
%For the Gogli data, this has to be one of {0,1,2,3} where 0 is the
%coarsest solution (128X128) that will not be annealed or refined in anyway
%
%low and high are the values for the soft threshold in the region-based
%distributing function
%
%path - this is the string that gives the path where the image is stored
%
%imgN - this is  the slice number of the image you want to segment
%
%Note - these values have been hardcoded in the script; they will have to
%be deleted before this can be used as a function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%Output%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%psi - multimask segmentation result (each color represents a different
%mask)
%TotTime - the approximate total time taken (in minutes)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
format compact;
TotTime=0;%measures the total time taken
epsilon=0.0001;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Initialization of the masks; more the randomness, greater the chance of
%achieving a 'good cut'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
L=4;
initP=2^(2*L); %Upto 1600 masks work without running into cache problems

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Initialization of the parameters used in the (force) computations - these
%values can be tuned further but segmentation performance does not depend
%critically on fine tuning these numbers
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
InScale=0.5;  %to dilate the error function used to compute the region-based distributing function
OutScale=0.9; %wt. of region-based distributing function
Kmax=1; %dilation factor, 0 for the corasest version
shrink=3-Kmax; %shrink=0 for native resolution

imgN=20; %this parameter is selected from {2,...,20} for Stack 4
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Creating a folder to store results along with the code and add the time stamp
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
current_time = datestr(now,'yyyy_mm_dd_HH_MM_SS');
filenamebase=sprintf('GolgiSeg_%s',current_time);
dos(sprintf('md %s',filenamebase));
dos(sprintf('copy main.m %s\\TIP_AMcode%s.m',filenamebase,num2str(imgN)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Reading an image from the stack
%Replace the path variable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
path = ('C:\08_SrinivasaFYLK_OriginalImages\Stack_4\');
if imgN<10
    img = double(imread(strcat(path,'YSG2-21_T0002_WavelengthA_Z000',num2str(imgN),'.tif')));
else
    img = double(imread(strcat(path,'YSG2-21_T0002_WavelengthA_Z00',num2str(imgN),'.tif')));
end

[MM,NN]=size(img);
%zero pad the image to avoid the masks from wrapping over
xPad=16;
yPad=16;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Shrink the image to the desired size b4 starting the computation and
%suitably pick the hue, saturation and variation of the various masks
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for l=1:shrink
    img=ComputeBlur(img);
end

[MM,NN]=size(img);
XX=zeros(MM,NN);
for m=1:MM/2-1
    XX(m+1,:)=m;
    XX(MM-m+1,:)=-m;
end
YY=zeros(MM,NN);
for n=1:NN/2-1;
    YY(:,n+1)=n;
    YY(:,NN-n+1)=-n;
end

S=0.5*ones(MM,NN);
V=img;
V = V-min(min(V));
V = V/(max(max(V)));
alpha = 0.5;
beta = 0.9;
V = (beta-alpha)*V+alpha;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Create an image mask (characteristic function for the image) and zero pad
%the original image and set the time counter on
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
imgMask = padarray(ones(size(img)),[xPad,yPad],0,'both');
img=padarray(img,[xPad,yPad],0,'both');
tic;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Define the dilation factors for annealing the region based force for
%different scales of the images (k) for different sizes (shrink) of the image
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for k=Kmax:-1:0
    if k==3
        dil=[24 22 20];
    elseif k==2
        dil=[20 18 16];
    elseif k==1
        dil=[16 14 12];
    elseif k==0
        dil=20; %Selected empirically based on size of the object to be segmented; other masks are scaled suitably
        %12 was found to create too sharp an image; ideally, the transition between
        %scale factors is not drastic
    end
    dil=(2^(-shrink))*dil; %If the dilation factors are not scaled with resolution then the image gets oversmoothed

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %For the dilation value at that scale of the image, compute the filter
    %that is used to compute the region based force at that scale and also
    %compute the model statistics of the image at this scale
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    for j=1:length(dil)
        a=dil(j);
        [Mpad,Npad]=size(img);
        [X,Y]=ndgrid([0:floor(Mpad/2),-floor(Mpad/2)+1:-1],[0:floor(Npad/2),-floor(Npad/2)+1:-1]);
        g1=exp(-(1/(a^2))*(X.^2+Y.^2)); % gaussian window
        g1Scale=ifft2(fft2(imgMask).*fft2(g1));
        f=(ifft2(fft2(img).*fft2(g1)))./(g1Scale+epsilon);
        f=f(xPad+1:end-xPad,yPad+1:end-yPad);
        low=2; %low and high values are picked based on rscalar
        high=4;

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %MR (blur and downsample) the image to the desired extent (size==k)
        %Initilize a MXNXP dimensional function, phi into which the contour is
        %embedded. The corresponding mask of interest is extracted in Psi.
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        for l=1:k
            f=ComputeBlur(f);
        end
        [M,N]=size(f); %used to compute the desired dimension of f

        fPad=padarray(ones(size(f)),[xPad,yPad],0,'both');

        if (k==Kmax) && (a==max(dil))%we have to intialize phi and psi for the very first run
            P=initP;
            colorP=P;
            colorIndex=1:P;
            phi=zeros(M,N,P); %multivalued function initlized at all zeros
            %%%%%%%%%%%%%%%%%%%% For perfect seeding use the following %%%%%%%%%%%%%%%%%%%%
%             load SW_Seeds_9.mat;
%             SW_Seeds=bwlabeln(SW_Seeds,4)+1;
%             psi=SW_Seeds(1:shrink:end,1:shrink:end);
            %%%%%%%%%%%%%%%%%%%% For random seeding use the following %%%%%%%%%%%%%%%%%%%%
            psi=ceil(P*rand(M,N)); %we introduce P numbers into the MXN mask Psi

            for p=1:P
                temp=phi(:,:,p);
                temp(psi==p)=1; %an indicator function to identify where psi has the value p
                phi(:,:,p)=temp; %the values in the pth dimension are set to the indicator function (above)
            end
            [maxval, psi] = max(phi,[],3); %the updated psi is computed as the maxarg of phi (in the pth dimension)
        elseif (k~=Kmax) && (a==max(dil)) % we continue working with the phi and psi that exist at the previous scale
            newPhi=zeros(M,N,P); %initialize a newPhi to match the new scale
            for p=1:P
                newPhi(:,:,p)=kron(phi(:,:,p),[1 1; 1 1]); %when the size of the image is doubled along each dimension
                %phi is dilated by dilating each of its values
            end
            clear phi;
            phi=newPhi;
            newPsi=kron(psi,[1 1; 1 1]); %values of psi are dilated in the same way
            clear psi;
            psi=newPsi;
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %Computation of the region-based distributing function is based on the smoothed 
        %version of the image; the error function is used to improve the classification 
        %of foreground and background
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        rscalar = -OutScale.*erf((2*f-low-high)/(InScale*(high-low))); 
        %rscalar is a coarse separation between foreground and
        %background regions with cusps between touching foreground objects

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %Computation of the Geometric (voting-based) distributing function with the scale b
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        fPad=padarray(ones(size(f)),[xPad,yPad],0,'both');
        [MPad,NPad]=size(fPad);
        fMask=padarray(ones(size(f)),[xPad,yPad],0,'both');
        [X,Y]=ndgrid([0:floor(MPad/2),-floor(MPad/2)+1:-1],[0:floor(NPad/2),-floor(NPad/2)+1:-1]);

        bfactor=1; %introduced to be able to weight the scale based on the size of the image (like 'shrink' is used with 'dil')
        b=(1/bfactor)*2^(6-k-shrink);%64 for the original image
        c=1/8; %this can be annealed along with b to figure out the 'best fit' for a scale value
        g2=0.5*(1-erf((X.^2+Y.^2-b^2)/(b^2+c))); %smaller the "support", slower the computation
        g2hat=fft2(g2);
        g2Scale=ifft2(fft2(fMask).*g2hat);

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %Keep track of how many pixels are changing; the algorithm stops when none of the pixels (in psi) move
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        change=1;
        iter=0;
        while change~=0
            tic;
            iter=iter+1;
            phiPad=padarray(phi,[xPad,yPad],'replicate','both');
            for p = 1:P
                phiPad(:,:,p) = (ifft2(fft2(phiPad(:,:,p)).*g2hat))./(g2Scale+epsilon);
                phi(:,:,p)=phiPad(xPad+1:end-xPad,yPad+1:end-yPad,p);
            end
            phi(:,:,1)=phi(:,:,1)+rscalar;
            [maxval, inwhat]=max(phi,[],3);
            newPsi=zeros(size(psi));
            phi=zeros(M,N,P);
            newP=0;
            for p=1:P %as phi and psi evolve, defragment phi to get rid of space allotted to nonexisting dimensions
                temp=zeros(size(psi));
                temp(inwhat==p)=1;
                if(sum(sum(temp))~=0) %check if any dimension of phi has been wiped out
                    newP=newP+1;
                    %%%%%%%%%% CHANGE THIS %%%%%%%%%%%%%%%%%%
                    phi(:,:,newP)=temp; %defragmenting the embedding function phi
                    colorIndex(newP)=colorIndex(p); %as we get rid of non existing dimension, we need to reassign color values to keep the hues from changing
                    newPsi = newPsi+colorIndex(p)*temp;
                end
            end
            P=newP;
            change=sum(sum(((newPsi-psi)~=0))); %how many pixels in the mask have moved?
            % The following three lines can be uncommented for a run time report of the algorithm's progress %
            %I=hsv2rgb(kron(mod((psi-1)/colorP+2/3,1),ones(2^k,2^k)),S,V); %Scale the coarse images up to equal the size of the detail image requested, so all the results are the same size
            %imwrite(I,sprintf('%s\\%02d_%02d_%03d.png',filenamebase,Kmax-k+1,j,iter));
            %disp(sprintf('k=%02d; a=%02d P=%03d; change=%05d; image %03d: Approximate time taken: %2.2f seconds',k,dil(j),P,change,iter,toc));
            TotTime = TotTime+toc; %this is a very approximate measure; comment the above three lines to get a better estimate
            psi=newPsi;
        end
    end
end
endTime=toc;
save(strcat(filenamebase,'\\endState.mat')); %this saves the workspace
beep;